@@ -2,66 +2,37 @@ import collections
 import re
 from copy import deepcopy
 from datetime import datetime
+from typing import Any, List, Optional, Union
+
 from . import scim_exceptions
 
 
-class AttributeFactory:
-
-    @staticmethod
-    def create(d, locator_path, attribute_type=None, is_parent_multi_valued=False, is_parent_complex=False):
-        locator_path = locator_path.copy() if isinstance(locator_path, list) else [locator_path]
-        multi_valued = d.get("multiValued", False)
-        if multi_valued and attribute_type is None:
-            return MultiValuedAttribute(
-                d=d,
-                locator_path=locator_path,
-                is_parent_multi_valued=is_parent_multi_valued,
-                is_parent_complex=is_parent_complex
-            )
-        attribute_type = d.get("type", "string") if attribute_type is None else attribute_type
-        attribute_factory = {
-            "binary": BinaryAttribute,
-            "boolean": BooleanAttribute,
-            "datetime": DatetimeAttribute,
-            "decimal": DecimalAttribute,
-            "complex": ComplexAttribute,
-            "integer": IntegerAttribute,
-            "reference": ReferenceAttribute,
-            "string": StringAttribute,
-        }
-        if attribute_type not in attribute_factory.keys():
-            raise AssertionError(
-                "Attribute type '{}' (path: {}) is not a valid type - expected on of these: ({})".format(
-                    attribute_type, locator_path, ", ".join(attribute_factory.keys())
-                )
-            )
-        attribute_class = attribute_factory[attribute_type]
-        return attribute_class(
-            d=d,
-            locator_path=locator_path,
-            is_parent_multi_valued=is_parent_multi_valued,
-            is_parent_complex=is_parent_complex
-        )
-
-
 class Attribute:
     _accepted_case_exact_value = {True, False}
     _accepted_uniqueness_value = {"none", "server", "global"}
 
-    def __init__(self, d, locator_path, is_parent_multi_valued=False, is_parent_complex=False):
+    def __init__(
+        self,
+        d: dict,
+        locator_path: Optional[List[str]],
+        is_parent_multi_valued: bool = False,
+        is_parent_complex: bool = False,
+        attribute_type: str = None,
+    ):
         # default values see - https://tools.ietf.org/html/rfc7643#section-2.2
         # Characteristics # https://tools.ietf.org/html/rfc7643#section-7
-        self.__d = d.copy() if not hasattr(self, "__d") is None else self.__d
+        self.__d: dict = d.copy() if not hasattr(self, "__d") is None else self.__d
         self._is_parent_multi_valued = is_parent_multi_valued
         self._is_parent_complex = is_parent_complex
 
         self.name = d.pop("name", None)
-        self._locator_path = locator_path
-        self._locator_path.append(self.name)
+        self._locator_path = locator_path.copy() if locator_path else []
+        if self.name:
+            self._locator_path.append(self.name)
 
         self.id = d.pop("id", None)
         # self.name = d.pop("name", self.name)
-        self.type = d.pop("type", "string")
+        self.type = attribute_type or d.pop("type")
         self.description = d.pop("description", None)
         self.required = d.pop("required", False)
         self.canonicalValues = d.pop("canonicalValues", None)
@@ -82,19 +53,21 @@ class Attribute:
                 property_name="id",
                 expected="Not None",
                 actual="None",
-                reference="https://tools.ietf.org/html/rfc7643#section-7"
+                reference="https://tools.ietf.org/html/rfc7643#section-7",
             )
 
-    def _validate_schema_name(self):
-        msg = "valid name e.g. must be ALPHA * {{nameChar}} where nameChar   = \"$\" / \"-\" / \"_\" / DIGIT / ALPHA"
+    def _validate_schema_name(self) -> None:
+        msg = 'valid name e.g. must be ALPHA * {{nameChar}} where nameChar   = "$" / "-" / "_" / DIGIT / ALPHA'
         # ^[a-zA-Z] - starts with ALPHA 0...many
         # (\$|\-|_|\w)$ - ends with $ - _ alphanumeric
-        if self.name is None or not bool(re.match('^[a-zA-Z](\$|-|_|\w)*$', self.name)):
+        if self.name is None or not bool(
+            re.match(r"^[a-zA-Z](\$|-|_|\w)*$", self.name)
+        ):
             raise scim_exceptions.ModelAttributeCharacteristicNotAllowedException(
                 locator_path=self._locator_path,
                 attribute_name="name",
                 expected=msg,
-                actual=self.name
+                actual=self.name,
             )
 
     def _validate_schema_required(self):
@@ -103,17 +76,18 @@ class Attribute:
                 locator_path=self._locator_path,
                 attribute_name="required",
                 expected="boolean",
-                actual=self.required
+                actual=self.required,
             )
 
     def _validate_schema_canonical_values(self):
-        if self.canonicalValues is not None and not isinstance(self.canonicalValues, list):
+        if self.canonicalValues is not None and not isinstance(
+            self.canonicalValues, list
+        ):
             raise scim_exceptions.ModelAttributeCharacteristicNotAllowedException(
                 locator_path=self._locator_path,
                 attribute_name="canonicalValues",
                 expected="none or valid list",
-                actual=self.canonicalValues
-
+                actual=self.canonicalValues,
             )
 
     def _validate_schema_case_exact(self):
@@ -122,7 +96,7 @@ class Attribute:
                 locator_path=self._locator_path,
                 attribute_name="caseExact",
                 expected=self._accepted_case_exact_value,
-                actual=self.caseExact
+                actual=self.caseExact,
             )
 
     def _validate_schema_mutability(self):
@@ -133,7 +107,7 @@ class Attribute:
                 locator_path=self._locator_path,
                 attribute_name="mutability",
                 expected=expected_values,
-                actual=self.mutability
+                actual=self.mutability,
             )
 
     def _validate_schema_returned(self):
@@ -143,7 +117,7 @@ class Attribute:
                 locator_path=self._locator_path,
                 attribute_name="returned",
                 expected=expected_values,
-                actual=self.returned
+                actual=self.returned,
             )
 
     def _validate_schema_uniqueness(self):
@@ -152,7 +126,7 @@ class Attribute:
                 locator_path=self._locator_path,
                 attribute_name="uniqueness_value",
                 expected=self._accepted_uniqueness_value,
-                actual=self.uniqueness
+                actual=self.uniqueness,
             )
 
     def _validate_schema_type(self):
@@ -161,7 +135,7 @@ class Attribute:
                 locator_path=self._locator_path,
                 attribute_name="type",
                 expected="a string",
-                actual=type(self.type)
+                actual=type(self.type),
             )
 
     # </editor-fold>
@@ -194,25 +168,33 @@ class Attribute:
                 scim_exceptions.ModelAttributeUnknownPropertyException(
                     attribute_name=self.name,
                     locator=self._locator_path,
-                    info="SCIM Schema parser does not recognise keys: '{}'".format(self._d.keys())
+                    info="SCIM Schema parser does not recognise keys: '{}'".format(
+                        self._d.keys()
+                    ),
                 )
             )
 
         if len(exceptions) > 0:
-            raise scim_exceptions.AggregatedScimSchemaExceptions(location=self._locator_path, exceptions=exceptions)
+            raise scim_exceptions.AggregatedScimSchemaExceptions(
+                location=self._locator_path, exceptions=exceptions
+            )
 
     @staticmethod
-    def _get_significant_value(d):
+    def _get_significant_value(d: dict) -> dict:
         return d
 
-    def _get_value(self, d):
+    def _get_value(self, d: dict) -> str:
         try:
             return d.pop(self.name)
         except KeyError:
-            raise scim_exceptions.ScimAttributeValueNotFoundException(d, self._locator_path, self.name, self.multiValued)
+            raise scim_exceptions.ScimAttributeValueNotFoundException(
+                d, self._locator_path, self.name, self.multiValued
+            )
 
-    def _validate(self, value):
-        raise NotImplementedError("Abstract class Attribute does not have validate method - use a concrete Class")
+    def _validate(self, value: Any):
+        raise NotImplementedError(
+            "Abstract class Attribute does not have validate method - use a concrete Class"
+        )
 
     def validate(self, d):
         try:
@@ -241,7 +223,7 @@ class BinaryAttribute(Attribute):
                 locator=self._locator_path,
                 value=value,
                 multi_value=self.multiValued,
-                attribute_type="binary"
+                attribute_type="binary",
             )
 
 
@@ -258,7 +240,7 @@ class BooleanAttribute(Attribute):
                 locator=self._locator_path,
                 value=value,
                 multi_value=self.multiValued,
-                attribute_type="boolean"
+                attribute_type="boolean",
             )
             # raise ValueError("{}-{} value: {} must be type boolean".format(self.id, self._locator_path, value))
 
@@ -269,7 +251,7 @@ class DatetimeAttribute(Attribute):
     _accepted_case_exact_value = {False}
     _accepted_uniqueness_value = {"none"}
 
-    def _validate(self, value):
+    def _validate(self, value: Any) -> None:
         try:
             datetime.strptime(value, "%Y-%m-%dT%H:%M:%SZ")
         except ValueError:
@@ -279,7 +261,7 @@ class DatetimeAttribute(Attribute):
                 value=value,
                 multi_value=self.multiValued,
                 attribute_type="datetime with format 2008-01-23T04:56:22Z",
-                reference=self._link_reference
+                reference=self._link_reference,
             )
 
 
@@ -288,12 +270,13 @@ class DecimalAttribute(Attribute):
     _link_reference = "https://tools.ietf.org/html/rfc7643#section-2.3.3"
     _accepted_case_exact_value = {False}
 
-    def _validate(self, value):
-        pos_period = str(value).index('.')
+    def _validate(self, value: Any) -> None:
+        pos_period = str(value).index(".")
         if not (
-                value and
-                isinstance(value, float) and
-                pos_period >= 1 and len(str(value)) - pos_period >= 1
+            value
+            and isinstance(value, float)
+            and pos_period >= 1
+            and len(str(value)) - pos_period >= 1
         ):
             type_description = "must be a real number with at least one digit to the left and right of the period"
             raise scim_exceptions.ScimAttributeInvalidTypeException(
@@ -311,14 +294,14 @@ class IntegerAttribute(Attribute):
     _link_reference = "https://tools.ietf.org/html/rfc7643#section-2.3.4"
     _accepted_case_exact_value = {True, False}
 
-    def _validate(self, value):
+    def _validate(self, value: Any) -> None:
         if not (not isinstance(value, bool) and isinstance(value, int)):
             raise scim_exceptions.ScimAttributeInvalidTypeException(
                 expected=self._d,
                 locator=self._locator_path,
                 value=value,
                 multi_value=self.multiValued,
-                attribute_type="integer"
+                attribute_type="integer",
             )
             # raise ValueError("{}-{} value: {} must be an integer".format(self.id, self._locator_path, value))
 
@@ -332,29 +315,31 @@ class ReferenceAttribute(Attribute):
     _accepted_case_exact_value = {True, False}
     referenceTypes = None
 
-    def _validate_schema_name(self):
+    def _validate_schema_name(self) -> None:
         if self.name == "$ref":
             return
 
         super()._validate_schema_name()
 
-    def __init__(self, d, locator_path, is_parent_multi_valued=False, is_parent_complex=False):
+    def __init__(
+        self, d, locator_path, is_parent_multi_valued=False, is_parent_complex=False
+    ):
         self.referenceTypes = d.pop("referenceTypes", None)
         super().__init__(
             d=d,
             locator_path=locator_path,
             is_parent_multi_valued=is_parent_multi_valued,
-            is_parent_complex=is_parent_complex
+            is_parent_complex=is_parent_complex,
         )
 
-    def _validate(self, value):
+    def _validate(self, value: Any) -> None:
         if not (isinstance(value, str)):
             raise scim_exceptions.ScimAttributeInvalidTypeException(
                 expected=self._d,
                 locator=self._locator_path,
                 value=value,
                 multi_value=self.multiValued,
-                attribute_type="type reference"
+                attribute_type="type reference",
             )
 
 
@@ -369,20 +354,25 @@ class StringAttribute(Attribute):
                 locator=self._locator_path,
                 value="({}){}".format(type(value).__name__, value),
                 multi_value=self.multiValued,
-                attribute_type="type string"
+                attribute_type="type string",
             )
 
         if self.canonicalValues:
             adjusted_value = value if self.caseExact else value.lower()
-            adjusted_canonical_value = self.canonicalValues if\
-                self.caseExact else [cv.lower() for cv in self.canonicalValues]
+            adjusted_canonical_value = (
+                self.canonicalValues
+                if self.caseExact
+                else [cv.lower() for cv in self.canonicalValues]
+            )
             if not (adjusted_value in adjusted_canonical_value):
                 raise scim_exceptions.ScimAttributeInvalidTypeException(
                     expected=self._d,
                     locator=self._locator_path,
                     value=value,
                     multi_value=self.multiValued,
-                    attribute_type="one of {}".format(" ,".join([v for v in adjusted_canonical_value]))
+                    attribute_type="one of {}".format(
+                        " ,".join([v for v in adjusted_canonical_value])
+                    ),
                 )
 
 
@@ -390,28 +380,33 @@ class ComplexAttribute(Attribute):
 
     _link_reference = "https://tools.ietf.org/html/rfc7643#section-2.3.8"
 
-    def __init__(self, d, locator_path, is_parent_multi_valued=False, is_parent_complex=False):
+    def __init__(
+        self, d, locator_path, is_parent_multi_valued=False, is_parent_complex=False
+    ):
         self.__d = d.copy()
         super().__init__(
             d=d,
             locator_path=locator_path,
             is_parent_multi_valued=is_parent_multi_valued,
-            is_parent_complex=is_parent_complex
+            is_parent_complex=is_parent_complex,
         )
         self.subAttributes = [
             AttributeFactory().create(
                 d=d,
                 locator_path=self._locator_path,
                 is_parent_multi_valued=is_parent_multi_valued,
-                is_parent_complex=True
-            ) for d in d.pop("subAttributes", [])
+                is_parent_complex=True,
+            )
+            for d in d.pop("subAttributes", [])
         ]
 
     def _get_significant_value(self, d):
         try:
             return d.get("value")
         except KeyError:
-            raise scim_exceptions.ScimAttributeValueNotFoundException(d, self._locator_path, self.name, self.multiValued)
+            raise scim_exceptions.ScimAttributeValueNotFoundException(
+                d, self._locator_path, self.name, self.multiValued
+            )
 
     def validate_schema(self):
         if self._is_parent_complex:
@@ -420,7 +415,7 @@ class ComplexAttribute(Attribute):
                 locator_path=self._locator_path,
                 attribute_name=self.name,
                 expected="simple type sub-attribute",
-                actual="complex type sub-attribute on a complex parent"
+                actual="complex type sub-attribute on a complex parent",
             )
 
         super().validate_schema()
@@ -432,7 +427,9 @@ class ComplexAttribute(Attribute):
                 exceptions.append(ae)
 
         if len(exceptions) > 0:
-            scim_exceptions.AggregatedScimSchemaExceptions(self._locator_path, exceptions=exceptions)
+            scim_exceptions.AggregatedScimSchemaExceptions(
+                self._locator_path, exceptions=exceptions
+            )
 
     def _validate(self, value):
         exceptions = []
@@ -447,8 +444,7 @@ class ComplexAttribute(Attribute):
 
         if len(exceptions) > 0:
             raise scim_exceptions.AggregatedScimMultValueAttributeValidationExceptions(
-                self._locator_path,
-                exceptions=exceptions
+                self._locator_path, exceptions=exceptions
             )
 
 
@@ -456,28 +452,46 @@ class MultiValuedAttribute(Attribute):
 
     _link_reference = "https://tools.ietf.org/html/rfc7643#section-2.4"
 
-    def __init__(self, d, locator_path, is_parent_multi_valued=False, is_parent_complex=False):
+    def __init__(
+        self,
+        d: dict,
+        locator_path: Optional[List[str]],
+        is_parent_multi_valued: bool = False,
+        is_parent_complex: bool = False,
+    ):
         self.__d = d.copy()
         self.type = d.get("type", None)  # shared attribute name to core
         self.primary = d.pop("primary", None)
         self.display = d.pop("display", None)
         self.value = d.pop("value", None)
         self.ref = d.pop("ref", None)
-        self.element_attribute = \
-            AttributeFactory().create(
-                d=d,
-                locator_path=locator_path,
-                attribute_type=self.type,
-                is_parent_multi_valued=True,
-                is_parent_complex=self.type == "Complex"
-            )
+        self.element_attribute = AttributeFactory().create(
+            d=d,
+            locator_path=locator_path,
+            attribute_type=self.type,
+            is_parent_multi_valued=True,
+            is_parent_complex=self.type == "Complex",
+        )
+
         super().__init__(
             d=d,
             locator_path=locator_path,
             is_parent_multi_valued=is_parent_multi_valued,
-            is_parent_complex=is_parent_complex
+            is_parent_complex=is_parent_complex,
+            attribute_type=self.type,
         )
 
+        self.name = self.element_attribute.name
+        self.multiValued = self.element_attribute.multiValued
+        self.id = self.element_attribute.id
+        self.description = self.element_attribute.description
+        self.required = self.element_attribute.required
+        self.canonicalValues = self.element_attribute.canonicalValues
+        self.caseExact = self.element_attribute.caseExact
+        self.mutability = self.element_attribute.mutability
+        self.returned = self.element_attribute.returned
+        self.uniqueness = self.element_attribute.uniqueness
+
     def _get_value(self, d):
         return self.element_attribute._get_value(d)
 
@@ -487,7 +501,7 @@ class MultiValuedAttribute(Attribute):
                 locator_path=self._locator_path,
                 attribute_name="type",
                 expected="a string",
-                actual=self.value
+                actual=self.value,
             )
 
     def _validate_schema_primary(self):
@@ -496,7 +510,7 @@ class MultiValuedAttribute(Attribute):
                 locator_path=self._locator_path,
                 attribute_name="primary",
                 expected="a boolean",
-                actual=self.primary
+                actual=self.primary,
             )
 
     def _validate_schema_display(self):
@@ -505,8 +519,8 @@ class MultiValuedAttribute(Attribute):
                 locator_path=self.display,
                 attribute_name="ref",
                 expected="a human-readable name (string)",
-                actual=self.display
-        )
+                actual=self.display,
+            )
 
     def _validate_schema_ref(self):
         if not isinstance(self.ref, str):
@@ -514,7 +528,7 @@ class MultiValuedAttribute(Attribute):
                 locator_path=self.display,
                 attribute_name="ref",
                 expected="a human-readable name (string)",
-                actual=self.ref
+                actual=self.ref,
             )
 
     def validate_schema(self):
@@ -522,19 +536,31 @@ class MultiValuedAttribute(Attribute):
 
     def _validate_uniqueness(self, list_values):
         if self.uniqueness:
-            difference_values = [item for item, count in collections.Counter(list_values).items() if count > 1]
+            difference_values = [
+                item
+                for item, count in collections.Counter(list_values).items()
+                if count > 1
+            ]
             if len(set(difference_values)) > 0:
-                raise scim_exceptions.ScimAttributeDuplicateValueException(locator=self._locator_path, value=difference_values)
+                raise scim_exceptions.ScimAttributeDuplicateValueException(
+                    locator=self._locator_path, value=difference_values
+                )
 
     def _validate(self, value):
         if not isinstance(value, list):
-            raise scim_exceptions.ScimAttributeInvalidTypeException(self._d, self._locator_path, value, self.multiValued, "list")
+            raise scim_exceptions.ScimAttributeInvalidTypeException(
+                self._d, self._locator_path, value, self.multiValued, "list"
+            )
 
         adjusted_values = [
-            self.element_attribute._get_significant_value(v) for v in value if (not v == {} and v is not None)
+            self.element_attribute._get_significant_value(v)
+            for v in value
+            if (not v == {} and v is not None)
         ]
         if len(adjusted_values) == 0:
-            raise scim_exceptions.ScimAttributeValueNotFoundException(value, self._locator_path, self.name, self.multiValued)
+            raise scim_exceptions.ScimAttributeValueNotFoundException(
+                value, self._locator_path, self.name, self.multiValued
+            )
 
         exceptions = []
         try:
@@ -552,5 +578,62 @@ class MultiValuedAttribute(Attribute):
         if len(exceptions) > 0:
             raise scim_exceptions.AggregatedScimMultValueAttributeValidationExceptions(
                 location="{} at path ('{}')".format(self.name, self._locator_path),
-                exceptions=exceptions
+                exceptions=exceptions,
             )
+
+
+attribute_factory = {
+    "binary": BinaryAttribute,
+    "boolean": BooleanAttribute,
+    "datetime": DatetimeAttribute,
+    "decimal": DecimalAttribute,
+    "complex": ComplexAttribute,
+    "integer": IntegerAttribute,
+    "reference": ReferenceAttribute,
+    "string": StringAttribute,
+}
+
+
+class AttributeFactory:
+    @staticmethod
+    def create(
+        d: dict,
+        locator_path: Union[str, Optional[List[str]]],
+        attribute_type: str = None,
+        is_parent_multi_valued: bool = False,
+        is_parent_complex: bool = False,
+    ):
+        if isinstance(locator_path, str):
+            locator_path = [locator_path]
+        elif isinstance(locator_path, list):
+            locator_path.copy()
+
+        multi_valued = d.get("multiValued", False)
+        if multi_valued and attribute_type is None:
+            return MultiValuedAttribute(
+                d=d,
+                locator_path=locator_path,
+                is_parent_multi_valued=is_parent_multi_valued,
+                is_parent_complex=is_parent_complex,
+            )
+
+        attribute_type = attribute_type or d.get("type")
+
+        if not (
+            attribute_type
+            and isinstance(attribute_type, str)
+            and attribute_type in attribute_factory.keys()
+        ):
+            raise AssertionError(
+                "Attribute type '{}' (path: {}) is not a valid type - expected on of these: ({})".format(
+                    attribute_type, locator_path, ", ".join(attribute_factory.keys())
+                )
+            )
+
+        attribute_class = attribute_factory[attribute_type]
+        return attribute_class(
+            d=d,
+            locator_path=locator_path,
+            is_parent_multi_valued=is_parent_multi_valued,
+            is_parent_complex=is_parent_complex,
+        )
